/****************************************************************************
** Meta object code from reading C++ file 'foreign_toplevel_manager_impl.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../src/modules/foreign-toplevel/impl/foreign_toplevel_manager_impl.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'foreign_toplevel_manager_impl.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t {};
} // unnamed namespace

template <> constexpr inline auto treeland_foreign_toplevel_manager_v1::qt_create_metaobjectdata<qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "treeland_foreign_toplevel_manager_v1",
        "before_destroy",
        "",
        "dockPreviewContextCreated",
        "treeland_dock_preview_context_v1*",
        "context"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'before_destroy'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'dockPreviewContextCreated'
        QtMocHelpers::SignalData<void(treeland_dock_preview_context_v1 *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<treeland_foreign_toplevel_manager_v1, qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject treeland_foreign_toplevel_manager_v1::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>.metaTypes,
    nullptr
} };

void treeland_foreign_toplevel_manager_v1::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<treeland_foreign_toplevel_manager_v1 *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->before_destroy(); break;
        case 1: _t->dockPreviewContextCreated((*reinterpret_cast<std::add_pointer_t<treeland_dock_preview_context_v1*>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< treeland_dock_preview_context_v1* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_manager_v1::*)()>(_a, &treeland_foreign_toplevel_manager_v1::before_destroy, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_manager_v1::*)(treeland_dock_preview_context_v1 * )>(_a, &treeland_foreign_toplevel_manager_v1::dockPreviewContextCreated, 1))
            return;
    }
}

const QMetaObject *treeland_foreign_toplevel_manager_v1::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *treeland_foreign_toplevel_manager_v1::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN36treeland_foreign_toplevel_manager_v1E_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int treeland_foreign_toplevel_manager_v1::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    return _id;
}

// SIGNAL 0
void treeland_foreign_toplevel_manager_v1::before_destroy()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void treeland_foreign_toplevel_manager_v1::dockPreviewContextCreated(treeland_dock_preview_context_v1 * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}
namespace {
struct qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t {};
} // unnamed namespace

template <> constexpr inline auto treeland_dock_preview_context_v1::qt_create_metaobjectdata<qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "treeland_dock_preview_context_v1",
        "before_destroy",
        "",
        "requestShow",
        "treeland_dock_preview_context_v1_preview_event*",
        "event",
        "requestShowTooltip",
        "treeland_dock_preview_tooltip_event*",
        "requestClose"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'before_destroy'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestShow'
        QtMocHelpers::SignalData<void(treeland_dock_preview_context_v1_preview_event *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'requestShowTooltip'
        QtMocHelpers::SignalData<void(treeland_dock_preview_tooltip_event *)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 5 },
        }}),
        // Signal 'requestClose'
        QtMocHelpers::SignalData<void()>(8, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<treeland_dock_preview_context_v1, qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject treeland_dock_preview_context_v1::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>.metaTypes,
    nullptr
} };

void treeland_dock_preview_context_v1::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<treeland_dock_preview_context_v1 *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->before_destroy(); break;
        case 1: _t->requestShow((*reinterpret_cast<std::add_pointer_t<treeland_dock_preview_context_v1_preview_event*>>(_a[1]))); break;
        case 2: _t->requestShowTooltip((*reinterpret_cast<std::add_pointer_t<treeland_dock_preview_tooltip_event*>>(_a[1]))); break;
        case 3: _t->requestClose(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (treeland_dock_preview_context_v1::*)()>(_a, &treeland_dock_preview_context_v1::before_destroy, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_dock_preview_context_v1::*)(treeland_dock_preview_context_v1_preview_event * )>(_a, &treeland_dock_preview_context_v1::requestShow, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_dock_preview_context_v1::*)(treeland_dock_preview_tooltip_event * )>(_a, &treeland_dock_preview_context_v1::requestShowTooltip, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_dock_preview_context_v1::*)()>(_a, &treeland_dock_preview_context_v1::requestClose, 3))
            return;
    }
}

const QMetaObject *treeland_dock_preview_context_v1::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *treeland_dock_preview_context_v1::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN32treeland_dock_preview_context_v1E_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int treeland_dock_preview_context_v1::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void treeland_dock_preview_context_v1::before_destroy()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void treeland_dock_preview_context_v1::requestShow(treeland_dock_preview_context_v1_preview_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void treeland_dock_preview_context_v1::requestShowTooltip(treeland_dock_preview_tooltip_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void treeland_dock_preview_context_v1::requestClose()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}
namespace {
struct qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t {};
} // unnamed namespace

template <> constexpr inline auto treeland_foreign_toplevel_handle_v1::qt_create_metaobjectdata<qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "treeland_foreign_toplevel_handle_v1",
        "before_destroy",
        "",
        "requestMaximize",
        "treeland_foreign_toplevel_handle_v1_maximized_event*",
        "event",
        "requestMinimize",
        "treeland_foreign_toplevel_handle_v1_minimized_event*",
        "requestActivate",
        "treeland_foreign_toplevel_handle_v1_activated_event*",
        "requestFullscreen",
        "treeland_foreign_toplevel_handle_v1_fullscreen_event*",
        "requestClose",
        "rectangleChanged",
        "treeland_foreign_toplevel_handle_v1_set_rectangle_event*",
        "State",
        "Maximized",
        "Minimized",
        "Activated",
        "Fullscreen"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'before_destroy'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestMaximize'
        QtMocHelpers::SignalData<void(treeland_foreign_toplevel_handle_v1_maximized_event *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'requestMinimize'
        QtMocHelpers::SignalData<void(treeland_foreign_toplevel_handle_v1_minimized_event *)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 5 },
        }}),
        // Signal 'requestActivate'
        QtMocHelpers::SignalData<void(treeland_foreign_toplevel_handle_v1_activated_event *)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 5 },
        }}),
        // Signal 'requestFullscreen'
        QtMocHelpers::SignalData<void(treeland_foreign_toplevel_handle_v1_fullscreen_event *)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 5 },
        }}),
        // Signal 'requestClose'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'rectangleChanged'
        QtMocHelpers::SignalData<void(treeland_foreign_toplevel_handle_v1_set_rectangle_event *)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 14, 5 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'State'
        QtMocHelpers::EnumData<enum State>(15, 15, QMC::EnumIsScoped).add({
            {   16, State::Maximized },
            {   17, State::Minimized },
            {   18, State::Activated },
            {   19, State::Fullscreen },
        }),
    };
    return QtMocHelpers::metaObjectData<treeland_foreign_toplevel_handle_v1, qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject treeland_foreign_toplevel_handle_v1::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>.metaTypes,
    nullptr
} };

void treeland_foreign_toplevel_handle_v1::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<treeland_foreign_toplevel_handle_v1 *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->before_destroy(); break;
        case 1: _t->requestMaximize((*reinterpret_cast<std::add_pointer_t<treeland_foreign_toplevel_handle_v1_maximized_event*>>(_a[1]))); break;
        case 2: _t->requestMinimize((*reinterpret_cast<std::add_pointer_t<treeland_foreign_toplevel_handle_v1_minimized_event*>>(_a[1]))); break;
        case 3: _t->requestActivate((*reinterpret_cast<std::add_pointer_t<treeland_foreign_toplevel_handle_v1_activated_event*>>(_a[1]))); break;
        case 4: _t->requestFullscreen((*reinterpret_cast<std::add_pointer_t<treeland_foreign_toplevel_handle_v1_fullscreen_event*>>(_a[1]))); break;
        case 5: _t->requestClose(); break;
        case 6: _t->rectangleChanged((*reinterpret_cast<std::add_pointer_t<treeland_foreign_toplevel_handle_v1_set_rectangle_event*>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)()>(_a, &treeland_foreign_toplevel_handle_v1::before_destroy, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)(treeland_foreign_toplevel_handle_v1_maximized_event * )>(_a, &treeland_foreign_toplevel_handle_v1::requestMaximize, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)(treeland_foreign_toplevel_handle_v1_minimized_event * )>(_a, &treeland_foreign_toplevel_handle_v1::requestMinimize, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)(treeland_foreign_toplevel_handle_v1_activated_event * )>(_a, &treeland_foreign_toplevel_handle_v1::requestActivate, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)(treeland_foreign_toplevel_handle_v1_fullscreen_event * )>(_a, &treeland_foreign_toplevel_handle_v1::requestFullscreen, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)()>(_a, &treeland_foreign_toplevel_handle_v1::requestClose, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (treeland_foreign_toplevel_handle_v1::*)(treeland_foreign_toplevel_handle_v1_set_rectangle_event * )>(_a, &treeland_foreign_toplevel_handle_v1::rectangleChanged, 6))
            return;
    }
}

const QMetaObject *treeland_foreign_toplevel_handle_v1::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *treeland_foreign_toplevel_handle_v1::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN35treeland_foreign_toplevel_handle_v1E_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int treeland_foreign_toplevel_handle_v1::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 7)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void treeland_foreign_toplevel_handle_v1::before_destroy()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void treeland_foreign_toplevel_handle_v1::requestMaximize(treeland_foreign_toplevel_handle_v1_maximized_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void treeland_foreign_toplevel_handle_v1::requestMinimize(treeland_foreign_toplevel_handle_v1_minimized_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void treeland_foreign_toplevel_handle_v1::requestActivate(treeland_foreign_toplevel_handle_v1_activated_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void treeland_foreign_toplevel_handle_v1::requestFullscreen(treeland_foreign_toplevel_handle_v1_fullscreen_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void treeland_foreign_toplevel_handle_v1::requestClose()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void treeland_foreign_toplevel_handle_v1::rectangleChanged(treeland_foreign_toplevel_handle_v1_set_rectangle_event * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}
QT_WARNING_POP
